%% 0.(a) Pre-processing and initialization
clear
clc

%Add paths, install dependencies if needed
addpath(genpath('./PlugPlay_v1/'));
addpath(genpath('./PlugPlay_v1/utilities/'));
addpath(genpath('./PlugPlay_v1/denoisers/BM3D/'));
addpath(genpath('./PlugPlay_v1/denoisers/TV/'));
addpath(genpath('./PlugPlay_v1/denoisers/NLM/'));
addpath(genpath('./PlugPlay_v1/denoisers/RF/'));
addpath(genpath('./BM3D/'));

mex vectors_nlmeans_single.c -v
mex image2vectors_single.c -v
mex vectors_nlmeans_double.c -v
mex image2vectors_double.c -v




%Load image, and make associated tag (all pictures of A will have the
%variable imageText in their file name.)
A = im2double(imread('Result Images/art_gray.png'));
imageText = 'art';

%i.e., if the image is RGB, convert to grayscale. The reason for grayscale
%is that I'm colorblind and find it quite hard to 
%discern differences in RGB.
if length(size(A)) > 2
    A = rgb2gray(A);
end

%If the image is too big, make it smaller. This is to make sure the
%computation time isn't too large. But the computation time is pretty large
%for the second section, so be warned.
if size(A,1) > 200
    A = imresize(A,0.5);
end

%define sizes (in pixels)
[M N P] = size(A);

%Define rho = 1.
rho = 1.0;

%% 0.(b) Masks
%Random Mask - The amount of pixels present is normally distributed
R = randn(size(A,1), size(A,2));
cdfNum = .45; %arbitrarily picked
randMask = abs(R)<=cdfNum; %~67% missing size P( abs(A) <= .45 ) ~ 1-.67

%Bar Mask - Make a mask where the only missing pixels are in a bar-shape
barWidthSize = 6; %units of pixels, width of bar
barMask = true .* ones(size(A,1), size(A,2)); %boolean mask
barMask(:,round(size(A,1)./2+1):round(size(A,1)./2+1)+barWidthSize) = false;
barMask(round(size(A,1)./2+1):round(size(A,1)./2+1)+barWidthSize,:) = false;
barMask = logical(barMask);

%Square Chunk Mask - Only missing pixels are in a square chunk
squareWidthSize = 55; %units of pixels


%In case I wanted to shift the square to anywhere.
shiftX = 10;
shiftY = -25;

squareMask = true .* ones(size(A,1), size(A,2));
squareMask(round(size(A,1)./2+1)-shiftX:round(size(A,1)./2+1)+squareWidthSize-shiftX,round(size(A,1)./2+1)-shiftY:round(size(A,1)./2+1)+squareWidthSize-shiftY) = false;
squareMask = logical(squareMask);


%Small Square Chunk Mask - Only missing pixels are in a (smaller) square chunk
squareWidthSize = 15; %units of pixels

smallsquareMask = true .* ones(size(A,1), size(A,2));
smallsquareMask(round(size(A,1)./2+1)-shiftX:round(size(A,1)./2+1)+squareWidthSize-shiftX,round(size(A,1)./2+1)-shiftY:round(size(A,1)./2+1)+squareWidthSize-shiftY) = false;
smallsquareMask = logical(smallsquareMask);


%% Section 1: Generating Images

count = 0;
methods = {'TV','NLM','BM3D', 'RF'}; %four priors
MaskNames = {'random', 'bar','square','smallsquare'}; %four masks

%Iterating through each prior
for ii = 1:length(methods)
    method = methods{ii};
    switch method
        case 'TV'
            lambda = 0.01;
        case 'NLM'
            lambda = 0.01;
        case 'BM3D'
            lambda = 0.005;
        case 'RF'
            lambda = 0.005;
    end
    
    %set a counter for the next for-loop
    counter = 0;
    
    %Iterating through each mask
    for MASK = {randMask, barMask, squareMask, smallsquareMask}
        counter = counter+1; %increment the counter by 1
        
        %overlay the mask on the uncorrupted image, yielding the different 
        %types of corruption
        AMask = A.*MASK{1}; 
       
        
        %Initialize the Plug-and-Play ADMM parameters
        opts.rho = rho;
        opts.gamma = 1;
        opts.maxitr = 50;
        opts.tol = 1e-4;
        
        %Implement the Image Inpainting using the Plug-and-Play ADMM
        Ul2 = PlugPlayADMM_inpaint(AMask,MASK{1},lambda,method,opts);

        %Calculate MSE/PSNR
        MSE = sum(sum( (A - Ul2).^2 )) ./ (size(A,1).*size(A,2));
        PSNR = 10.*log10(max(max(A)).^2 ./ MSE);

        %Calculate SSIM. SSIM has its own function!
        SSIM = ssim(Ul2,A);
        
        %Display results to console
        disp(sprintf('The PSNR of the inpainted image is %.2f dB', PSNR))
        disp(sprintf('The SSIM of the inpainted image is %.4f', SSIM))

        %Plot images: original image, corrupted image, and inpainted image
        figure()
        subplot(131)
        imshow(A)
        title(sprintf('Original Image'),'HorizontalAlignment', 'center','Interpreter','LaTex')

        subplot(132)
        imshow(AMask)
        if MASK{1} == randMask
            DataMissing = 1 - (normcdf(cdfNum) - normcdf(-cdfNum));
            title(sprintf('Corrupted Image\n %.03f Percent of Data Missing',...
                100.*DataMissing),'HorizontalAlignment', 'center','Interpreter','LaTex')
        else
            title(sprintf('Corrupted Image'),'HorizontalAlignment',...
                'center','Interpreter','LaTex')
        end

        subplot(133)
        imshow(Ul2)

        text1 = sprintf(['Recovered Image, ', method ,...
            ' Prior\n $\\lambda$ = %.03f, $\\rho$ = %.02f\n PSNR = %0.2f dB, SSIM = %0.4f'],...
            lambda, rho, PSNR, SSIM);
        title(text1,'HorizontalAlignment', 'center','Interpreter','LaTex')
        print(gcf, '-djpeg', [imageText, '_', lower(method),'_',...
            MaskNames{counter}])
    end
end
%% Analysis of PSNR/SSIM vs. \lambda
%NOTE: Don't run this section if you want to see results fairly quickly --
%this section is takes quite a lot of time to run.
%initialize figure, vector of PSNRs/SSIMs for each prior
figure()
L = 100;
PSNRVecs = zeros(L,4);
SSIMVecs = zeros(L,4);

%iterate through each prior
for ii = 1:length(methods)
    method = methods{ii};
    PSNRVec = [];
    SSIMVec = [];
    
    %Compute image inpainting for values of lambda between 0.001 and 0.1
    for lambda = linspace(0.001,0.1,L)
        AMask = A.*randMask;
        
        %Initialize the Plug-and-Play ADMM parameters
        opts.rho = rho;
        opts.gamma = 1;
        opts.maxitr = 50;
        opts.tol = 1e-4;
        
        %Implement the Image Inpainting using the Plug-and-Play ADMM
        Ul2 = PlugPlayADMM_inpaint(AMask,randMask,lambda,method,opts);

        %Calculate MSE/PSNR
        MSE = sum(sum( (A - Ul2).^2 )) ./ (size(A,1).*size(A,2));
        PSNR = 10.*log10(max(max(A)).^2 ./ MSE);

        %Calculate SSIM
        SSIM = ssim(Ul2,A);

        %Save values in vector
        PSNRVec = [PSNRVec; PSNR];
        SSIMVec = [SSIMVec; SSIM];
        
        %Display to console
        disp(sprintf('The PSNR of the inpainted image is %.2f dB', PSNR))
        disp(sprintf('The SSIM of the inpainted image is %.4f', SSIM))
        
    end
    %Add vector of values to matrix for plotting
    PSNRVecs(:,ii) = PSNRVec;
    SSIMVecs(:,ii) = SSIMVec;
end


%Plotting
figure()
plot(linspace(0.001,0.1,100),PSNRVecs)
legend('TV','NLM','BM3D', 'RF')
xlabel('$\lambda$','Interpreter','LaTex')
ylabel('PSNR [dB]','Interpreter','LaTex')
title('PSNR vs. $\lambda$','Interpreter','LaTex')
figure()
plot(linspace(0.001,0.1,100),SSIMVecs)
xlabel('$\lambda$','Interpreter','LaTex')
ylabel('SSIM','Interpreter','LaTex')
title('SSIM vs. $\lambda$','Interpreter','LaTex')
legend('TV','NLM','BM3D', 'RF')

